from math import pi, cos
from numpy import empty
from .latticesection import LatticeSection
from .latticepanel import LatticePanel
from .latticestrip import LatticeStrip
from pygeom.geom3d import ihat, Point, Coordinate

class LatticeSheet(object):
    sect1 = None
    sect2 = None
    bspace = None
    yspace = None
    mirror = None
    levec = None
    cord = None
    nrml = None
    strps = None
    pnls = None
    mind = None
    ctrls = None
    def __init__(self, sect1: LatticeSection, sect2: LatticeSection):
        self.sect1 = sect1
        self.sect2 = sect2
        self.update()
    def update(self):
        if self.sect1.mirror or self.sect2.mirror:
            self.mirror = True
        else:
            self.mirror = False
        self.bspace = inherit_spacing(self)
        self.ctrls = inherit_controls(self)
        self.levec = self.sect2.pnt-self.sect1.pnt
        vecz = (ihat**self.levec).to_unit()
        vecy = (vecz**ihat).to_unit()
        self.cord = Coordinate(self.sect1.pnt, ihat, vecy, vecz)
        self.strps = []
        self.pnls = []
    def mesh_strips(self, lsid: int):
        self.strps = []
        pnta = self.sect1.pnt
        vecr = self.levec
        crda = self.sect1.chord
        crdb = self.sect2.chord
        crdr = crdb-crda
        anga = self.sect1.angle
        angb = self.sect2.angle
        angr = angb-anga
        lenb = len(self.bspace)
        for i in range(lenb):
            bspc = self.bspace[i]
            bsp1 = bspc[0]
            bsp2 = bspc[2]
            pnt1 = pnta+bsp1*vecr
            pnt2 = pnta+bsp2*vecr
            crd1 = crda+bsp1*crdr
            crd2 = crda+bsp2*crdr
            strp = LatticeStrip(lsid, pnt1, pnt2, crd1, crd2, bspc)
            ang1 = anga+bsp1*angr
            ang2 = anga+bsp2*angr
            strp.set_angles(ang1, ang2)
            strp.sht = self
            self.strps.append(strp)
            lsid += 1
        return lsid
    def inherit_panels(self):
        self.pnls = []
        for strp in self.strps:
            for pnl in strp.pnls:
                self.pnls.append(pnl)
    def set_control_panels(self):
        # print(f'# Sheet Panels = {len(self.pnls):d}')
        for control in self.ctrls:
            # print(f'Setting {control:s} panels!')
            ctrl = self.ctrls[control]
            for pnl in self.pnls:
                if pnl.cspc[3] >= ctrl.xhinge:
                    ctrl.add_panel(pnl)
    # def set_control_points_and_normals(self):
    #     for pnl in self.pnls:
    #         pntc = pnl.return_panel_point()
    #         pnl.set_control_point(pntc)
    #         pnl.set_camber_angle(0.0)
    def __repr__(self):
        return '<LatticeSheet>'

def inherit_spacing(sht: LatticeSheet):
    if sht.mirror:
        if sht.sect2.bspace is None:
            bspace = [(0.0, 0.5, 1.0)]
        else:
            lenb = len(sht.sect2.bspace)
            bspace = []
            for i in range(lenb):
                blst = []
                for j in range(2, -1, -1):
                    blst.append(1.0-sht.sect2.bspace[i][j])
                bspace.append(tuple(blst))
            bspace.reverse()
    else:
        if sht.sect1.bspace is None:
            bspace = [(0.0, 0.5, 1.0)]
        else:
            bspace = sht.sect1.bspace
    return bspace

def inherit_controls(sht: LatticeSheet):
    ctrls = {}
    if sht.mirror:
        for control in sht.sect2.ctrls:
            ctrl = sht.sect2.ctrls[control]
            newctrl = ctrl.duplicate(mirror=True)
            ctrls[control] = newctrl
    else:
        for control in sht.sect1.ctrls:
            ctrl = sht.sect1.ctrls[control]
            newctrl = ctrl.duplicate(mirror=False)
            ctrls[control] = newctrl
    for control in ctrls:
        ctrl = ctrls[control]
        if ctrl.uhvec.return_magnitude() == 0.0:
            pnt1 = sht.sect1.pnt
            crd1 = sht.sect1.chord
            pnta = pnt1+crd1*ihat*ctrl.xhinge
            pnt2 = sht.sect2.pnt
            crd2 = sht.sect2.chord
            pntb = pnt2+crd2*ihat*ctrl.xhinge
            hvec = pntb-pnta
            ctrl.set_hinge_vector(hvec)
    return ctrls



from .latticesystem import LatticeSystem, latticesystem_from_json
from .latticesurface import LatticeSurface
from .latticesection import LatticeSection
from .latticesheet import LatticeSheet
from .latticestrip import LatticeStrip
from .latticepanel import LatticePanel
from .latticeresult import LatticeResult
from .latticeoptimum import LatticeOptimum
from .latticecontrol import LatticeControl

cfrm = '.5f'
dfrm = '.7f'
efrm = '.4f'



from math import log, pi, copysign

class FlatPlate(object):
    def __init__(self):
        pass
    def return_camber(self, xc: float):
        return 0.0
    def return_camber_slope(self, xc: float):
        return 0.0
    def return_camber_angle(self, xc: float):
        return 0.0
    def __repr__(self):
        return f'<Flat Plate Airfoil>'

class NACA4(object):
    code = None
    _mt = None
    _mc = None
    _pc = None
    def __init__(self, code: str):
        self.code = code
    @property
    def mt(self):
        if self._mt is None:
            self._mt = float(self.code[2])/10+float(self.code[3])/100
        return self._mt
    @property
    def mc(self):
        if self._mc is None:
            self._mc = float(self.code[0])/100
        return self._mc
    @property
    def pc(self):
        if self._pc is None:
            self._pc = float(self.code[1])/10
        return self._pc
    def return_camber(self, xc: float):
        if xc < self.pc:
            yc = self.mc/self.pc**2*(2*self.pc*xc-xc**2)
        else:
            yc = self.mc/(1-self.pc)**2*((1-2*self.pc)+2*self.pc*xc-xc**2)
        return yc
    def return_camber_slope(self, xc: float):
        if xc < self.pc:
            dydx = self.mc/self.pc**2*(2*self.pc-2*xc)
        else:
            dydx = self.mc/(1-self.pc)**2*(2*self.pc-2*xc)
        return dydx
    def return_camber_angle(self, xc: float):
        from math import degrees, atan
        dydx = self.return_camber_slope(xc)
        return degrees(atan(dydx))
    def __repr__(self):
        return f'<NACA {self.code:s}>'

class NACA6Series(object):
    code = None
    a = None
    xx = None
    q = None
    cl = None
    h0 = None
    h = None
    g = None
    cff = None
    oma = None
    def __init__(self, code: str):
        if code[0] != '6':
            print('Not a NACA 6 series code.')
            return None
        a = float(code[1])/10
        xx = float(code[-2:])/100
        q = float(code[-3])
        clstr = code[2:-3].lstrip('(').rstrip(')')
        cl = float(clstr)/10
        self.code = code
        self.a = a
        self.xx = xx
        self.q = q
        self.cl = cl
        g = (a**2*(0.5*log(a) - 0.25) + 0.25)/(a - 1)
        self.g = g
        h0 = (0.25 - 0.5*log(1 - a))*(a - 1)
        self.h = h0 + g
        self.cff = cl/(2*pi*(a+1))
    def return_camber(self, xc: float):
        if xc == 0.0:
            yc = 0.0
        else:
            cl = self.cl
            a = self.a
            g = self.g
            h = self.h
            yc = cl*((a - 1)*(g - h*xc - xc*log(xc)) - 0.5*(a - xc)**2*log(abs(a - xc)) + 0.25*(a - xc)**2 + 0.5*(xc - 1)**2*log(1 - xc) - 0.25*(xc - 1)**2)/(2*pi*(a - 1)*(a + 1))
        return yc
    def return_camber_slope(self, xc: float):
        cl = self.cl
        a = self.a
        h = self.h
        cs = cl/(2*pi*(a + 1))
        if xc == 0.0:
            dydx = float('inf')
        elif xc == a:
            dydx = cs*(log(1 - a) - log(a) - h - 1)
        elif xc == 1.0:
            dydx = cs*(copysign((a - xc)**2/2, a - xc) + ((xc-a)/2 - (a - 1)*(h + log(xc) + 1) + (a - xc)*log(abs(a - xc)))*abs(a - xc))/((a - 1)*abs(a - xc))
        else:
            dydx = cs*(copysign((a - xc)**2/2, a - xc) + ((xc-a)/2 - (a - 1)*(h + log(xc) + 1) + (a - xc)*log(abs(a - xc)) + (xc - 1)*log(1 - xc))*abs(a - xc))/((a - 1)*abs(a - xc))
        return dydx
    def return_camber_angle(self, xc: float):
        from math import degrees, atan
        dydx = self.return_camber_slope(xc)
        return degrees(atan(dydx))
    def __repr__(self):
        return f'<NACA {self.code:s}>'



from pygeom.geom3d import Vector
# from .latticesheet import LatticeSheet

class LatticeControl(object):
    name = None
    posgain = None
    neggain = None
    xhinge = None
    uhvec = None
    reverse = None
    pnls = None
    def __init__(self, name: str, posgain: float, neggain: float, xhinge: float):
        self.name = name
        self.posgain = posgain
        self.neggain = neggain
        self.xhinge = xhinge
        self.pnls = []
    def set_hinge_vector(self, hvec: Vector):
        if hvec.return_magnitude() != 0.0:
            self.uhvec = hvec.to_unit()
        else:
            self.uhvec = hvec
    def duplicate(self, mirror=True):
        if mirror and self.reverse:
            posgain, neggain = -self.neggain, -self.posgain
        else:
            posgain, neggain = self.posgain, self.neggain
        if mirror:
            uhvec = Vector(self.uhvec.x, -self.uhvec.y, self.uhvec.z)
        else:
            uhvec = Vector(self.uhvec.x, self.uhvec.y, self.uhvec.z)
        ctrl = LatticeControl(self.name, posgain, neggain, self.xhinge)
        ctrl.reverse = False
        ctrl.set_hinge_vector(uhvec)
        return ctrl
    def add_panel(self, pnl):
        self.pnls.append(pnl)

def latticecontrol_from_json(name: str, controldata: dict):
    xhinge = controldata['xhinge']
    posgain = 1.0
    if 'posgain' in controldata:
        posgain = controldata['posgain']
    neggain = 1.0
    if 'neggain' in controldata:
        neggain = controldata['neggain']
    ctrl = LatticeControl(name, posgain, neggain, xhinge)
    hvec = Vector(0.0, 0.0, 0.0)
    if 'hvec' in controldata:
        hvec = vector_from_json(controldata['hvec'])
    ctrl.set_hinge_vector(hvec)
    reverse = False
    if 'reverse' in controldata:
        reverse = controldata['reverse']
    ctrl.reverse = reverse    
    return ctrl

def vector_from_json(vectordata):
    x, y, z = None, None, None
    if 'x' in vectordata:
        x = vectordata['x']
    if 'y' in vectordata:
        y = vectordata['y']
    if 'z' in vectordata:
        z = vectordata['z']
    return Vector(x, y, z)



from math import asin, cos, pi
from pygeom.geom3d import Point, ihat
from pyvlm.tools.camber import FlatPlate
from .latticecontrol import LatticeControl, latticecontrol_from_json

class LatticeSection(object):
    pnt = None
    chord = None
    angle = None
    camber = None
    bspace = None
    yspace = None
    mirror = None
    ctrls = None
    def __init__(self, pnt: Point, chord: float, angle: float):
        self.pnt = pnt
        self.chord = chord
        self.angle = angle
        self.update()
    def update(self):
        self.mirror = False
        self.camber = FlatPlate()
        self.ctrls = {}
    def set_span_equal_spacing(self, numb: int):
        from pyvlm.tools import equal_spacing
        bsp = equal_spacing(2*numb)
        self.bspace = [tuple(bsp[i*2:i*2+3]) for i in range(numb)]
    def set_span_cosine_spacing(self, numb: int):
        from pyvlm.tools import full_cosine_spacing
        bsp = full_cosine_spacing(2*numb)
        self.bspace = [tuple(bsp[i*2:i*2+3]) for i in range(numb)]
    def set_span_semi_cosine_spacing(self, numb: int):
        from pyvlm.tools import semi_cosine_spacing
        bsp = semi_cosine_spacing(2*numb)
        self.bspace = [tuple(bsp[i*2:i*2+3]) for i in range(numb)]
    def set_airfoil(self, airfoil: str):
        if airfoil[0:4].lower() == 'naca':
            code = airfoil[4:].strip()
            if len(code) == 4:
                from ..tools.camber import NACA4
                self.camber = NACA4(code)
                self.airfoil = airfoil
            elif code[0] == '6':
                from ..tools.camber import NACA6Series
                self.camber = NACA6Series(code)
                self.airfoil = airfoil
    def add_control(self, ctrl: LatticeControl):
        self.ctrls[ctrl.name] = ctrl
    def return_mirror(self):
        pnt = Point(self.pnt.x, -self.pnt.y, self.pnt.z)
        chord = self.chord
        angle = self.angle
        sect = LatticeSection(pnt, chord, angle)
        sect.camber = self.camber
        sect.bspace = self.bspace
        sect.yspace = self.yspace
        sect.mirror = True
        sect.ctrls = self.ctrls
        return sect
    def return_point(self, percrd: float):
        return self.pnt+self.chord*percrd*ihat
    def get_camber(self, xc: float):
        return self.camber.cubic_interp(xc)
    def __repr__(self):
        return '<LatticeSection>'

def latticesecttion_from_json(sectdata: dict):
    xle = sectdata['xle']
    yle = sectdata['yle']
    zle = sectdata['zle']
    crd = sectdata['chord']
    if 'angle' in sectdata:
        ang = sectdata['angle']
    else:
        ang = 0.0
    pnt = Point(xle, yle, zle)
    sect = LatticeSection(pnt, crd, ang)
    if 'airfoil' in sectdata:
        sect.set_airfoil(sectdata['airfoil'])
    if 'numb' in sectdata and 'bspace' in sectdata:
        numb = sectdata['numb']
        bspace = sectdata['bspace']
        if bspace == 'equal':
            sect.set_span_equal_spacing(numb)
        elif bspace == 'cosine':
            sect.set_span_cosine_spacing(numb)
        elif bspace == 'semi-cosine':
            sect.set_span_semi_cosine_spacing(numb)
    if 'controls' in sectdata:
        for name in sectdata['controls']:
            ctrl = latticecontrol_from_json(name, sectdata['controls'][name])
            sect.add_control(ctrl)
    return sect
